
#include "config.h"
#include "stack.h"

static __thread int ass_integer;
static __thread int ass_memory;

static long ass_regs[] = {
	LT_64_REG_RDI,
	LT_64_REG_RSI,
	LT_64_REG_RDX,
	LT_64_REG_RCX,
	LT_64_REG_R8,
	LT_64_REG_R9,
};
#define ASS_REGS_CNT  (sizeof(ass_regs)/sizeof(int))
#define ASS_REGS_SIZE (sizeof(ass_regs))


static int classificate_memory(struct lt_config_shared *cfg, struct lt_arg *arg, int align)
{
	if (align)
		ass_memory = LT_STACK_ALIGN(ass_memory);

	PRINT_VERBOSE2(cfg->verbose, 2, "%s - ass %d\n", 
			arg->name, ass_memory);

	ARCH_SET(arg, ARCH_FLAG_MEM, ass_memory);

	if (arg->pointer)
		ass_memory += sizeof(void*);
	else
		ass_memory += arg->type_len;
		
	return 0;
}

static int classificate_integer(struct lt_config_shared *cfg, struct lt_arg *arg, int align)
{
	if (align)
		ass_integer = LT_STACK_ALIGN(ass_integer);
	else {
		/* We need to find out if we fit into the 
		   register with the rgument type width,
		   if not we will move to next register */
		int ass = LT_STACK_ALIGN(ass_integer);

		if ((ass_integer + arg->type_len) > ass)
			ass_integer = LT_STACK_ALIGN(ass_integer);
	}

	PRINT_VERBOSE2(cfg->verbose, 2, "ass %s %d %d\n", 
			arg->name, ASS_REGS_SIZE, ass_integer);

	if (ASS_REGS_SIZE == ass_integer)
		return -1;
	
	ARCH_SET(arg, ARCH_FLAG_REG, ass_integer);

	if (arg->pointer)
		ass_integer += sizeof(void*);
	else
		ass_integer += arg->type_len;

	return 0;
}

static void struct_arch(struct lt_config_shared *cfg, struct lt_arg *sarg, struct lt_arg *farg)
{
	if (sarg->pointer)
		return;

	/* Structures passed by value dont have the arch assigned by default.
	   If the first argument is returned in memory, the structure takes
	   this pointer, if it is register, we keep it NULL, and lt_args_cb_arg
	    will print out the REG kw for value.*/
	PRINT_VERBOSE2(cfg->verbose, 2, "first argument for struct %s has flag: %d\n", 
			sarg->name, ARCH_GET_FLAG(farg));

	if (ARCH_GET_FLAG(farg) == ARCH_FLAG_MEM)
		sarg->arch = farg->arch;
	else
		ARCH_SET(sarg, ARCH_FLAG_SVREG, 0);
}

static int classificate_struct(struct lt_config_shared *cfg, struct lt_arg *arg)
{
	struct lt_arg *a;
	int allmem = 0;
	int first = 1;

	PRINT_VERBOSE2(cfg->verbose, 2, "struct %s\n", arg->name);

	/* If the structure is bigger than 16B then 
	   it is in the memory (no registers). */
	if (arg->type_len > 16)
		allmem = 1;

	lt_list_for_each_entry(a, arg->args_head, args_list) {

		if (!allmem) {
			if (!classificate_integer(cfg, a, first)) {

				if (first)
					struct_arch(cfg, arg, a);

				first = 0;
				continue;
			}

			/* Once there is no register left, 
			   we are going to memory. */
		}

		classificate_memory(cfg, a, first);

		if (first)
			struct_arch(cfg, arg, a);

		first = 0;
	}

	return 0;
}

static int classificate_arg(struct lt_config_shared *cfg, struct lt_arg *arg)
{
	int class;

	do {
		/* pointers are INTEGER class by default */
		if (arg->pointer) {
			class = LT_CLASS_INTEGER;
			break;
		}

		switch(arg->type_id) {
		case LT_ARGS_TYPEID_CHAR:
		case LT_ARGS_TYPEID_UCHAR:
		case LT_ARGS_TYPEID_SHORT:
		case LT_ARGS_TYPEID_USHORT:
		case LT_ARGS_TYPEID_INT:
		case LT_ARGS_TYPEID_UINT:
		case LT_ARGS_TYPEID_LONG:
		case LT_ARGS_TYPEID_ULONG:
			class = LT_CLASS_INTEGER;
			break;
		default:
			class = LT_CLASS_NONE;
			break;
		}

	} while(0);

	PRINT_VERBOSE2(cfg->verbose, 2, "argument %s dtype %d - class %d\n", 
			arg->name, arg->dtype, class);
	return class;
}

static int classificate(struct lt_config_shared *cfg, struct lt_args_sym *sym)
{
	int i;

	ass_integer = 0;
	ass_memory  = 0;

	PRINT_VERBOSE2(cfg->verbose, 2, "got symbol \"%s\"\n", sym->name);

	for(i = 1; i < sym->argcnt; i++) {

		int class;
		struct lt_arg *arg;

		arg = sym->args[i];

		if ((arg->dtype != LT_ARGS_DTYPE_POD) && 
		    (!arg->pointer)) {
			if (-1 == classificate_struct(cfg, arg))
				return -1;
			continue;
		}

		class = classificate_arg(cfg, arg);
		if (-1 == class)
			return -1;

		/* XXX Strange so far... might be needed in future 
 		   when more type classes are added. */
		switch(class) {
		case LT_CLASS_INTEGER:
			if (-1 != classificate_integer(cfg, arg, 1))
				break;
			/* Falling throught in case we ran out of registers. */

		case LT_CLASS_MEMORY:
			classificate_memory(cfg, arg, 1);
			break;
		}
	}

	return 0;
}

static void *get_value_mem(struct lt_config_shared *cfg, struct lt_arg *arg, La_regs *regs, int ret_stval)
{
	long offset = ARCH_GET_OFFSET(arg);

	void *pval = regs->lr_rsp +                         /* current stack pointer */
		     sizeof(void*) +                        /* return function address */
		     ( ret_stval ? sizeof(void*) : 0 ) +    /* return address of structure returned by value */
		     (void*) (offset);

	PRINT_VERBOSE2(cfg->verbose, 2, "offset = %ld, %s = %p\n", offset, arg->name, pval);
	return pval;
}

static void *get_value_reg(struct lt_config_shared *cfg, struct lt_arg *arg, La_regs *regs)
{
	void *pval = NULL;

	long offset = ARCH_GET_OFFSET(arg);
	long aoff   = LT_STACK_ALIGN(offset);
	long qoff   = aoff - offset;
	long reg    = ass_regs[offset / sizeof(long)];

	PRINT_VERBOSE2(cfg->verbose, 2, "offset %ld - aoff %ld - reg %ld - ASS_REGS_CNT %ld\n", 
			offset, aoff, reg, ASS_REGS_CNT);

	switch(reg) {
	case LT_64_REG_RDI:
		pval = &regs->lr_rdi;
		break;
	case LT_64_REG_RSI:
		pval = &regs->lr_rsi;
		break;
	case LT_64_REG_RDX:
		pval = &regs->lr_rdx;
		break;
	case LT_64_REG_RCX:
		pval = &regs->lr_rcx;
		break;
	case LT_64_REG_R8:
		pval = &regs->lr_r8;
		break;
	case LT_64_REG_R9:
		pval = &regs->lr_r9;
		break;
	}

	pval += qoff;

	PRINT_VERBOSE2(cfg->verbose, 2, "argument %s = %p\n", arg->name, pval);
	return pval;
}

static void* get_value(struct lt_config_shared *cfg, struct lt_arg *arg, La_regs *regs, int ret_stval)
{
	void *pval = NULL;

	PRINT_VERBOSE2(cfg->verbose, 2, "get value for %s - arch = %lx\n", 
			arg->name, ARCH_GET(arg));

	if (ARCH_FLAG_MEM == ARCH_GET_FLAG(arg))
		pval = get_value_mem(cfg, arg, regs, ret_stval);
	else if (ARCH_FLAG_REG == ARCH_GET_FLAG(arg))
		pval = get_value_reg(cfg, arg, regs);

	return pval;
}

/* Process structure stored completelly in the 
   memory - pointed to by 'pval' arg. */
static int process_struct_mem(struct lt_config_shared *cfg, struct lt_arg *arg, void *pval, struct lt_args_data *data)
{
	struct lt_arg *a;
	int i = 0;

	PRINT_VERBOSE2(cfg->verbose, 2, "for %s - arch = %llx\n", 
			arg->name, ARCH_GET(arg));

	pval += ARCH_GET_OFFSET(arg);

	lt_args_cb_struct(cfg, LT_ARGS_STRUCT_ITSELF, arg, NULL, data, 0);

	lt_list_for_each_entry(a, arg->args_head, args_list) {
		int last = (i + 1) == arg->mmbcnt;

		lt_args_cb_struct(cfg, LT_ARGS_STRUCT_ARG, a, pval, data, last);

		pval += a->type_len;
		i++;
	}

	return 0;
}

/* Process classified structure - could 
   be in registers or memory. */
static int process_struct_arg(struct lt_config_shared *cfg, struct lt_arg *arg, La_regs *regs, int ret_stval, struct lt_args_data *data)
{
	struct lt_arg *a;
	int i = 0;

	lt_args_cb_struct(cfg, LT_ARGS_STRUCT_ITSELF, arg, NULL, data, 0);

	lt_list_for_each_entry(a, arg->args_head, args_list) {
		int last = (i + 1) == arg->mmbcnt;
		void *pval = NULL;

		pval = get_value(cfg, a, regs, ret_stval);

		lt_args_cb_struct(cfg, LT_ARGS_STRUCT_ARG, a, pval, data, last);
		i++;
	}

	return 0;
}

int lt_stack_process(struct lt_config_shared *cfg, struct lt_args_sym *asym, La_regs *regs, struct lt_args_data *data)
{
	int i;
	struct lt_arg *rarg = asym->args[LT_ARGS_RET];
	int ret_stval = (!rarg->pointer) &&
			(rarg->dtype >= LT_ARGS_TYPEID_CUSTOM);

	for(i = 1; i < asym->argcnt; i++) {

		struct lt_arg *arg = asym->args[i];
		int last = (i + 1) == asym->argcnt;
		void *pval = NULL;

		if (!ARCH_GET(arg) &&
		    (-1 == classificate(cfg, asym)))
			return -1;

		pval = get_value(cfg, arg, regs, ret_stval);
		lt_args_cb_arg(cfg, arg, pval, data, last, 1);

		if ((cfg->args_detailed) &&
		    (LT_ARGS_DTYPE_STRUCT == arg->dtype)) {

			if (arg->pointer) {

				if (!pval) {
					PRINT_VERBOSE2(cfg->verbose, 2, "THIS should never happen - arg '%s %s'\n",
							arg->type_name, arg->name);
					continue;
				}

				pval = *((void**) pval);
				process_struct_mem(cfg, arg, pval, data);

			} else 
				process_struct_arg(cfg, arg, regs, ret_stval, data);
		}
	}

	return 0;
}

int lt_stack_process_ret(struct lt_config_shared *cfg, struct lt_args_sym *asym, La_retval *regs, struct lt_args_data *data)
{
	struct lt_arg *arg;

	void *pval = &regs->lrv_rax;
	arg = asym->args[LT_ARGS_RET];

	lt_args_cb_arg(cfg, arg, pval, data, 1, 0);

	if ((cfg->args_detailed) &&
	    (LT_ARGS_DTYPE_STRUCT == arg->dtype)) {

		/* The process_struct_mem expects the pointer to the structure,
 		   so in case the argument is pointer to the structure the 
		   pointer we need to return the pointer itself */
		if (arg->pointer)
                        pval = (void*) regs->lrv_rax;

		process_struct_mem(cfg, arg, pval, data);
	}

	return 0;
}
